from typing import Optional

import click
import pandas as pd

from taxonomical_utils.merger import merge_files
from taxonomical_utils.resolver import resolve_taxa
from taxonomical_utils.upper_taxa_lineage_appender import append_upper_taxa_lineage
from taxonomical_utils.wikidata_fetcher import wd_taxo_fetcher_from_ott


@click.group()
def cli() -> None:
    pass


@click.command(name="resolve_taxa")
@click.option("--input-file", required=True, type=click.Path(exists=True), help="Path to the input file.")
@click.option("--output-file", required=True, type=click.Path(), help="Path to the output file.")
@click.option("--org-column-header", required=True, type=str, help="Column header for the organism.")
def resolve_taxa_cli(input_file: str, output_file: str, org_column_header: str) -> None:
    resolve_taxa(input_file, output_file, org_column_header)


@click.command(name="append_taxonomy")
@click.option("--input-file", required=True, type=click.Path(exists=True), help="Path to the input file.")
@click.option("--output-file", required=True, type=click.Path(), help="Path to the output file.")
def append_taxonomy_cli(input_file: str, output_file: str) -> None:
    append_upper_taxa_lineage(input_file, output_file)


@click.command(name="append_wd_id")
@click.option("--input-file", required=True, type=click.Path(exists=True), help="Path to the input file.")
@click.option("--output-file", required=True, type=click.Path(), help="Path to the output file.")
def append_wd_id_cli(input_file: str, output_file: str) -> None:
    url = "https://query.wikidata.org/sparql"
    input_df = pd.read_csv(input_file)
    results = []
    for ott_id in input_df["ott_id"]:
        result_df = wd_taxo_fetcher_from_ott(url, ott_id)
        results.append(result_df)
    final_df = pd.concat(results, ignore_index=True)
    final_df.to_csv(output_file, index=False)


@click.command(name="merge")
@click.option("--input-file", required=True, type=click.Path(exists=True), help="Path to the input file.")
@click.option("--resolved-taxa-file", type=click.Path(exists=True), help="Path to the resolved taxa file.")
@click.option("--upper-taxa-lineage-file", type=click.Path(exists=True), help="Path to the upper taxa lineage file.")
@click.option("--wd-file", type=click.Path(exists=True), help="Path to the Wikidata ID file.")
@click.option("--output-file", required=True, type=click.Path(), help="Path to the output file.")
@click.option("--org-column-header", required=True, type=str, help="Column header for the organism.")
@click.option("--delimiter", default=",", type=str, help="Delimiter of the input file.")
@click.option("--remove-intermediate", is_flag=True, help="Remove intermediate files.")
def merge_cli(
    input_file: str,
    output_file: str,
    org_column_header: str,
    delimiter: str,
    resolved_taxa_file: Optional[str] = None,
    upper_taxa_lineage_file: Optional[str] = None,
    wd_file: Optional[str] = None,
    remove_intermediate: bool = False,
) -> None:
    merge_files(
        input_file,
        output_file,
        org_column_header,
        delimiter,
        resolved_taxa_file,
        upper_taxa_lineage_file,
        wd_file,
        remove_intermediate,
    )


cli.add_command(resolve_taxa_cli)
cli.add_command(append_taxonomy_cli)
cli.add_command(append_wd_id_cli)
cli.add_command(merge_cli)

if __name__ == "__main__":
    cli()



import json
from typing import Any, Dict, List

import pandas as pd
from opentree import OT
from pandas import json_normalize


# Define the utility functions
def save_json(data: Any, filename: str) -> None:
    with open(filename, "w") as f:
        json.dump(data, f)


def load_json(filename: str) -> Any:
    with open(filename) as f:
        return json.load(f)


def convert_ott_ids_to_int(df: pd.DataFrame, column_name: str) -> pd.DataFrame:
    df[column_name] = df[column_name].astype("Int64")
    return df


def get_unique_ott_ids(df: pd.DataFrame, column_name: str) -> List[int]:
    return list(df[column_name].dropna().astype("int"))


def fetch_taxon_info(ott_ids: List[int]) -> List[Dict[str, Any]]:
    taxon_info = []
    for ott_id in ott_ids:
        query = OT.taxon_info(ott_id, include_lineage=True)
        taxon_info.append(query.response_dict)
    return taxon_info


def extract_lineage(json_data: List[Dict[str, Any]]) -> pd.DataFrame:
    # Process the JSON data
    lineage_data = []

    for entry in json_data:
        if "lineage" in entry and entry["lineage"]:
            lineage_data.append(entry)
        else:
            # Create a fake lineage for entries with an empty lineage
            fake_lineage = [
                {
                    "name": entry["unique_name"],
                    "ott_id": entry["ott_id"],
                    "flags": entry["flags"],
                    "rank": entry["rank"],
                }
            ]
            lineage_data.append({**entry, "lineage": fake_lineage})

    # Normalize the JSON data
    df = json_normalize(
        lineage_data,
        record_path=["lineage"],
        meta=["ott_id", "unique_name", "flags", "rank"],
        record_prefix="sub_",
        errors="ignore",
    )

    return df


def pivot_taxonomy(df: pd.DataFrame, value_column: str, suffix: str = "") -> pd.DataFrame:
    pivot_df = df.pivot(index="ott_id", columns="sub_rank", values=value_column)
    if suffix:
        pivot_df = pivot_df.add_suffix(suffix)
    return pivot_df


def merge_taxonomy_dataframes(df1: pd.DataFrame, df2: pd.DataFrame) -> pd.DataFrame:
    return pd.merge(df1, df2, how="left", on="ott_id")


def add_lowest_taxon_rank(df: pd.DataFrame) -> pd.DataFrame:
    for index, row in df.iterrows():
        rank = row["rank"]
        unique_name_col = f"{rank}"
        ott_id_col = f"{rank}_ott_id"
        df.at[index, unique_name_col] = row["unique_name"]
        df.at[index, ott_id_col] = row["ott_id"]
    return df


def rename_columns(df: pd.DataFrame, renaming_dict: Dict[str, str]) -> pd.DataFrame:
    return df.rename(columns=renaming_dict)


# Main function to append upper taxonomy lineage
def append_upper_taxa_lineage(input_file: str, output_file: str) -> pd.DataFrame:
    # Load the merged dataframe from the input file
    merged_df = pd.read_csv(input_file, sep=",", encoding="utf-8")

    # Convert ott_ids to int
    merged_df = convert_ott_ids_to_int(merged_df, "taxon.ott_id")
    ott_list = get_unique_ott_ids(merged_df, "taxon.ott_id")

    # Fetch taxon info
    taxon_info = fetch_taxon_info(ott_list)

    # Save taxon info to JSON
    # We keep the original input file name, strip the extension and add "_upper_taxa_lineage.json" to it

    input_file_no_ext = input_file.split(".")[0]

    taxon_info_filename = f"{input_file_no_ext}_upper_taxa_lineage.json"
    save_json(taxon_info, taxon_info_filename)

    # Load and normalize JSON
    json_data = load_json(taxon_info_filename)

    df_tax_lineage_filtered = extract_lineage(json_data)

    # Filter and pivot taxonomy data
    df_tax_lineage_filtered = df_tax_lineage_filtered.groupby(["ott_id", "sub_rank"], as_index=False).last()
    df_tax_lineage_filtered_flat = pivot_taxonomy(df_tax_lineage_filtered, "sub_name")
    df_tax_lineage_filtered_flat_ott_ids = pivot_taxonomy(df_tax_lineage_filtered, "sub_ott_id", "_ott_id")

    # Add species name and remove duplicates
    df_tax_lineage_filtered_flat = merge_taxonomy_dataframes(
        df_tax_lineage_filtered_flat, df_tax_lineage_filtered[["ott_id", "unique_name"]]
    )
    df_tax_lineage_filtered_flat.drop_duplicates(subset=["ott_id", "unique_name"], inplace=True)

    # Merge taxonomy IDs
    df_tax_lineage_filtered_flat_full = merge_taxonomy_dataframes(
        df_tax_lineage_filtered_flat, df_tax_lineage_filtered_flat_ott_ids
    )

    # Drop duplicates and handle flags and ranks
    df_tax_lineage_filtered["flags_tuple"] = df_tax_lineage_filtered["flags"].apply(lambda x: tuple(x))
    df_tax_lineage_filtered_sub = df_tax_lineage_filtered.drop_duplicates(
        subset=["ott_id", "flags_tuple", "rank"], keep="first", ignore_index=False
    )

    # Merge with ranks
    merged = merge_taxonomy_dataframes(
        df_tax_lineage_filtered_flat_full, df_tax_lineage_filtered_sub[["ott_id", "flags", "rank"]]
    )
    merged = add_lowest_taxon_rank(merged)

    # Rename columns
    renaming_dict = {
        "kingdom": "organism_otol_kingdom",
        "phylum": "organism_otol_phylum",
        "class": "organism_otol_class",
        "order": "organism_otol_order",
        "family": "organism_otol_family",
        "genus": "organism_otol_genus",
        "species": "organism_otol_species",
        "unique_name": "organism_otol_unique_name",
        "kingdom_ott_id": "organism_otol_kingdom_ott_id",
        "phylum_ott_id": "organism_otol_phylum_ott_id",
        "class_ott_id": "organism_otol_class_ott_id",
        "order_ott_id": "organism_otol_order_ott_id",
        "family_ott_id": "organism_otol_family_ott_id",
        "genus_ott_id": "organism_otol_genus_ott_id",
        "species_ott_id": "organism_otol_species_ott_id",
    }
    merged = rename_columns(merged, renaming_dict)

    # Ensure all necessary columns are present
    required_columns = [
        "ott_id",
        "organism_otol_kingdom",
        "organism_otol_phylum",
        "organism_otol_class",
        "organism_otol_order",
        "organism_otol_family",
        "organism_otol_genus",
        "organism_otol_species",
        "organism_otol_unique_name",
        "organism_otol_kingdom_ott_id",
        "organism_otol_phylum_ott_id",
        "organism_otol_class_ott_id",
        "organism_otol_order_ott_id",
        "organism_otol_family_ott_id",
        "organism_otol_genus_ott_id",
        "organism_otol_species_ott_id",
    ]
    for column in required_columns:
        if (column not in merged.columns) or (merged[column].isnull().all()):
            merged[column] = pd.NA

    # Reorder columns as listed above and keep any additional columns
    merged = merged[required_columns + [col for col in merged.columns if col not in required_columns]]

    # Save to CSV
    merged.to_csv(output_file, sep=",", index=False, encoding="utf-8")

    return merged



from typing import Optional

import pandas as pd

from taxonomical_utils.processor import process_species_list
from taxonomical_utils.wikidata_fetcher import wd_taxo_fetcher_from_ott

url = "https://query.wikidata.org/sparql"
ott_id = 514975  # Example ott_id
wd_taxo_fetcher_from_ott(url, ott_id)

input_file = "tests/data/sample_data.csv"
output_file = "tests/data/merged_output.csv"
org_column_header = "idTaxon"
resolved_taxa_file = "tests/data/sample_data_treated.csv"
upper_taxa_lineage_file = "tests/data/sample_data_upper_taxo.csv"
wd_file = "tests/data/sample_data_wd.csv"
delimiter = ","


def merge_files(
    input_file: str,
    output_file: str,
    org_column_header: str,
    delimiter: str = ",",
    resolved_taxa_file: Optional[str] = None,
    upper_taxa_lineage_file: Optional[str] = None,
    wd_file: Optional[str] = None,
) -> None:
    input_df = pd.read_csv(input_file, delimiter=delimiter)

    # Process the species list in the input file
    input_df = process_species_list(input_file, org_column_header, delimiter)

    if resolved_taxa_file:
        resolved_taxa_df = pd.read_csv(resolved_taxa_file)
        resolved_taxa_df = resolved_taxa_df.add_prefix("otl_")
        input_df = pd.merge(
            input_df, resolved_taxa_df, left_on="taxon_search_string", right_on="otl_search_string", how="left"
        )

    if upper_taxa_lineage_file and "otl_taxon.ott_id" in input_df.columns:
        upper_taxa_lineage_df = pd.read_csv(upper_taxa_lineage_file)
        upper_taxa_lineage_df = upper_taxa_lineage_df.add_prefix("otl_")
        input_df = pd.merge(
            input_df, upper_taxa_lineage_df, left_on="otl_taxon.ott_id", right_on="otl_ott_id", how="left"
        )

    if wd_file and "otl_taxon.ott_id" in input_df.columns:
        wd_df = pd.read_csv(wd_file)
        wd_df = wd_df.add_prefix("wd_")
        input_df = pd.merge(input_df, wd_df, left_on="otl_taxon.ott_id", right_on="wd_ott.value", how="left")

    input_df.to_csv(output_file, sep=delimiter, index=False)


merge_files(
    input_file=input_file,
    output_file=output_file,
    org_column_header=org_column_header,
    delimiter=delimiter,
    resolved_taxa_file=resolved_taxa_file,
    upper_taxa_lineage_file=upper_taxa_lineage_file,
    wd_file=wd_file,
)



import json
from typing import Any, Dict, List, Tuple

import pandas as pd


def save_json(data: Any, filepath: str) -> None:
    with open(filepath, "w") as out:
        json.dump(data, out, indent=2, sort_keys=True)


def load_json(filepath: str) -> Any:
    with open(filepath) as tmpfile:
        return json.load(tmpfile)


def normalize_json_upper(json_data: List[Dict[str, Any]]) -> Tuple[pd.DataFrame, pd.DataFrame]:
    matches = pd.json_normalize(json_data, record_path=["results", "matches"])
    unmatched = pd.DataFrame(json_data[0]["unmatched_names"], columns=["unmatched_names"])
    return matches, unmatched


def normalize_json_resolver(json_data: Dict[str, Any]) -> Tuple[pd.DataFrame, pd.DataFrame]:
    matches = pd.json_normalize(json_data, record_path=["results", "matches"])
    unmatched = pd.DataFrame(json_data["unmatched_names"], columns=["unmatched_names"])
    return matches, unmatched



import pandas as pd
import requests
from pandas import json_normalize


def wd_taxo_fetcher_from_ott(url: str, ott_id: int) -> pd.DataFrame:
    query = f"""
    PREFIX wdt: <http://www.wikidata.org/prop/direct/>
    SELECT ?ott ?wd ?img
    WHERE{{
        ?wd wdt:P9157 ?ott
        OPTIONAL{{ ?wd wdt:P18 ?img }}
        VALUES ?ott {{'{ott_id}'}}
    }}
    """

    r = requests.get(url, params={"format": "json", "query": query}, timeout=10)
    data = r.json()
    results = pd.DataFrame.from_dict(data).results.bindings
    df = json_normalize(results)

    # Handle duplicates by keeping only the first occurrence for each 'ott.value'
    df = df.drop_duplicates(subset=["ott.value"], keep="first")

    return df



import os

import pandas as pd

from taxonomical_utils.exceptions import FileDownloadError
from taxonomical_utils.processor import process_species_list, resolve_organisms
from taxonomical_utils.shared import load_json, normalize_json_resolver, save_json


def resolve_taxa(
    input_file: str,
    output_file: str,
    org_column_header: str,
) -> pd.DataFrame:
    # Define paths
    path_to_input_file = input_file

    # Check if the input file exists
    if not os.path.isfile(path_to_input_file):
        raise FileDownloadError(path_to_file=path_to_input_file)

    # Detect file type by extension if delimiter is not specified
    delimiter = "," if input_file.endswith(".csv") else "\t"

    # Process species list
    species_list_df = process_species_list(path_to_input_file, org_column_header, delimiter)

    # Resolve organisms
    organisms = species_list_df["taxon_search_string"].unique().tolist()
    # Here also we make sure to remove NaN values
    organisms = [x for x in organisms if str(x) != "nan"]
    print(f"Resolving {len(organisms)} organisms")
    print(organisms)
    organisms_tnrs_matched = resolve_organisms(organisms)

    # Save taxon info to JSON
    # We keep the original input file name, strip the extension and add "_taxon_info.json" to it

    input_file_no_ext = input_file.split(".")[0]

    taxon_info_filename = f"{input_file_no_ext}_taxon_info.json"
    save_json(organisms_tnrs_matched, taxon_info_filename)

    # Load and normalize json
    json_data = load_json(taxon_info_filename)
    df_organism_tnrs_matched, df_organism_tnrs_unmatched = normalize_json_resolver(json_data)

    if not df_organism_tnrs_matched.empty:
        # Process the results and update the dataframe
        df_organism_tnrs_matched.sort_values(["search_string", "is_synonym"], axis=0, inplace=True)

        # Ensure we keep all unique matches for each search_string
        df_organism_tnrs_matched.drop_duplicates(subset=["search_string"], keep="first", inplace=True)

        # Drop duplicates based on the provided org_column_header
        df_organism_tnrs_matched.drop_duplicates(
            subset=["search_string", "matched_name", "taxon.ott_id"], keep="first", inplace=True
        )

        # Save the final dataframe
        df_organism_tnrs_matched.to_csv(output_file, sep=",", index=False, encoding="utf-8")

    else:
        print(f"No organisms were resolved for {organisms}")

    return df_organism_tnrs_matched
